#include <hdf5.h>
#include <mpi.h>
#include <stdio.h>
#include <stdlib.h>
#include <assert.h>
#include <unistd.h>
#include <malloc.h>

double heap_mallinfo(void)
{
  // unsigned int mmap, arena, small;
  // double heap;
  // struct mallinfo mi; /* structure in bytes */

  // mi = mallinfo();
  // mmap = mi.hblkhd;
  // arena = mi.uordblks;
  // small = mi.usmblks;
  // heap = ((double)(mmap + arena + small))/1024.0; /* convert to KB */
  // return heap;
  struct mallinfo minfo = mallinfo();
  double used = (double) ((unsigned int) minfo.hblkhd + 0.0 + 
                          (unsigned int) minfo.usmblks + 
                          (unsigned int) minfo.uordblks);
  /* This is in bytes, we need KB */
  return used/1024.0;
}

#define gpts 100
#define nbands  2000

/* Simple test for writing wafefunction-like data with HDF5
   Usage: mpirun -np 8 ./a.out 2 2 2 2
   The first parameter is the state parallelization, the following
   denote domain decomposition. Parallelization options have to be 
   compatible with the total number of processors */

int main(int argc, char *argv[]) {

  int my_id, nprocs;
  int mpi_dims[4];
  int period[4] = {0, 0, 0, 0};
  int coords[4];

  hid_t file_id, filespace_id, memspace_id, dset_id, plist_id;
  hsize_t dimsf[4] = {nbands, gpts, gpts, gpts};
  hsize_t count[4];
  hsize_t offset[4];
  hsize_t *maxdims = NULL;
  int ndims = 4;

  double t0, t1, t2, t3;
  double mem1, mem2, mem3, mem4;
  double mem1_max, mem2_max, mem3_max, mem4_max;
  double mem1_ave, mem2_ave, mem3_ave, mem4_ave;
  double *my_data;

  MPI_Comm cart_comm;

  MPI_Init(&argc, &argv);
  MPI_Comm_size(MPI_COMM_WORLD, &nprocs);

  assert(argc == 5);
  for (int i=1; i < argc; i++)
    mpi_dims[i-1] = atoi(argv[i]);

  assert(mpi_dims[0] * mpi_dims[1] * mpi_dims[2] * mpi_dims[3] == nprocs);
  MPI_Cart_create(MPI_COMM_WORLD, 4, mpi_dims, period, 0, &cart_comm);
  MPI_Comm_rank(cart_comm, &my_id);

  MPI_Cart_coords(cart_comm, my_id, 4, coords);

  assert(nbands % mpi_dims[0] == 0);
  for (int i=1; i < 4; i++)
    assert(gpts % mpi_dims[i] == 0);

  int total_size = nbands*gpts*gpts*gpts;
  count[0] = nbands / mpi_dims[0];
  offset[0] = coords[0] * count[0];
  int data_size = count[0];
  for (int i=1; i < 4; i++)
    {
      count[i] = gpts/mpi_dims[i];
      offset[i] = coords[i] * count[i];
      data_size *= count[i];
    }

  if (my_id == 0) fprintf(stderr, "malloc psi\n");
  my_data = (double *) malloc(data_size * sizeof(double));
  for (int i=0; i < data_size; i++)
    my_data[i] = my_id;

  MPI_Barrier(MPI_COMM_WORLD);

  /* Reference memory */
  mem1 = heap_mallinfo();
  MPI_Reduce(&mem1, &mem1_max, 1, MPI_DOUBLE, MPI_MAX, 0, cart_comm);
  MPI_Reduce(&mem1, &mem1_ave, 1, MPI_DOUBLE, MPI_SUM, 0, cart_comm);
  mem1_ave /= nprocs;

  t0 = MPI_Wtime();

  if (my_id == 0) fprintf(stderr, "H5Pcreate\n");
  plist_id = H5Pcreate(H5P_FILE_ACCESS);
  if (my_id == 0) fprintf(stderr, "H5Pset\n");
  H5Pset_fapl_mpio(plist_id, cart_comm, MPI_INFO_NULL);
  if (my_id == 0) fprintf(stderr, "H5Fcreate\n");
  file_id = H5Fcreate("test.dat", H5F_ACC_TRUNC, H5P_DEFAULT, plist_id);
  H5Pclose(plist_id);

  if (my_id == 0) fprintf(stderr, "H5Screate_simple\n");
  filespace_id = H5Screate_simple(ndims, dimsf, NULL);
  if (my_id == 0) fprintf(stderr, "H5Dcreate\n");
  dset_id = H5Dcreate(file_id, "testdata", H5T_NATIVE_DOUBLE, filespace_id, 
		      H5P_DEFAULT, H5P_DEFAULT, H5P_DEFAULT); 
  H5Sclose(filespace_id);

  if (my_id == 0) fprintf(stderr, "H5Screate_simple\n");
  memspace_id = H5Screate_simple(ndims, count, NULL);
  if (my_id == 0) fprintf(stderr, "H5Dget_space\n");
  filespace_id = H5Dget_space(dset_id);
  if (my_id == 0) fprintf(stderr, "H5Sselect_hyperslab\n");
  H5Sselect_hyperslab(filespace_id, H5S_SELECT_SET, offset, NULL, count, NULL);
  
  if (my_id == 0) fprintf(stderr, "H5Pcreate\n");
  plist_id = H5Pcreate(H5P_DATASET_XFER);
  H5Pset_dxpl_mpio(plist_id, H5FD_MPIO_COLLECTIVE);

  /* Memory before write */
  mem2 = heap_mallinfo();
  MPI_Reduce(&mem2, &mem2_max, 1, MPI_DOUBLE, MPI_MAX, 0, cart_comm);
  MPI_Reduce(&mem2, &mem2_ave, 1, MPI_DOUBLE, MPI_SUM, 0, cart_comm);
  mem2_ave /= nprocs;

  /* Do write */
#ifdef TAU
  TAU_START("H5Dwrite");
#endif
  if (my_id == 0) fprintf(stderr, "H5Dwrite (before) \n");
  H5Dwrite(dset_id, H5T_NATIVE_DOUBLE, memspace_id, filespace_id,
	   plist_id, my_data);
  if (my_id == 0) fprintf(stderr, "H5Dwrite (after) \n");
#ifdef TAU
  TAU_STOP("H5Dwrite");
#endif

  /* Memory after write */
  mem3 = heap_mallinfo();
  MPI_Reduce(&mem3, &mem3_max, 1, MPI_DOUBLE, MPI_MAX, 0, cart_comm);
  MPI_Reduce(&mem3, &mem3_ave, 1, MPI_DOUBLE, MPI_SUM, 0, cart_comm);
  mem3_ave /= nprocs;
  
  H5Dclose(dset_id);
  H5Sclose(filespace_id);
  H5Sclose(memspace_id);
  H5Pclose(plist_id);
  H5Fclose(file_id);
  MPI_Barrier(MPI_COMM_WORLD);

  /* Memory after close */
  mem4 = heap_mallinfo();
  MPI_Reduce(&mem4, &mem4_max, 1, MPI_DOUBLE, MPI_MAX, 0, cart_comm);
  MPI_Reduce(&mem4, &mem4_ave, 1, MPI_DOUBLE, MPI_SUM, 0, cart_comm);
  mem4_ave /= nprocs;

  t1 = MPI_Wtime();

  if (my_id == 0)
    {
      printf("IO time %f (%f) MB %f s\n",
             total_size * 8/(1024.0*1024.0),
             data_size * 8/(1024.0*1024.0), t1-t0);

      printf("Memory rank = 0: %f %f %f %f\n", mem1, mem2, mem3, mem4);
      printf("Memory usage max (ave): %f (%f) %f (%f) %f (%f) %f (%f)\n",
	     mem1_max, mem1_ave, mem2_max, mem2_ave, mem3_max, mem3_ave, mem4_max, mem4_ave);

    }

  MPI_Finalize();
}
